@@ -23,7 +23,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <popt.h>
-#include <sys/time.h> /* TODO: replace timeofday with g_timers */
 #include <gdk/gdkwindow.h>
 #include <gtk/gtk.h>
 #ifdef USE_GDKPIXBUF_RENDER_TO_DRAWABLE
@@ -63,7 +62,9 @@ enum {
 	ZOOM_REGION_SMOOTHSCROLL_PROP,
 	ZOOM_REGION_INVERT_PROP,
 	ZOOM_REGION_SMOOTHING_PROP,
-	ZOOM_REGION_CONTRAST_PROP,
+	ZOOM_REGION_CONTRASTR_PROP,
+	ZOOM_REGION_CONTRASTG_PROP,
+	ZOOM_REGION_CONTRASTB_PROP,
 	ZOOM_REGION_XSCALE_PROP,
 	ZOOM_REGION_YSCALE_PROP,
 	ZOOM_REGION_BORDERSIZE_PROP,
@@ -85,7 +86,9 @@ gchar* prop_names[ZOOM_REGION_EXIT_MAGNIFIER + 1] =
     "SMOOTHSCROLL",
     "INVERT",
     "SMOOTHING",
-    "CONTRAST",
+    "CONTRASTR",
+    "CONTRASTG",
+    "CONTRASTB",
     "XSCALE",
     "YSCALE",
     "BORDERSIZE",
@@ -107,82 +110,6 @@ typedef enum {
 	ZOOM_REGION_ERROR_TOO_BIG
 } ZoomRegionPixmapCreationError;
 
-/*
- * When running the magnifier in timing test mode (--timing-pan-rate and/or --timing-iterations),
- * then the following output is displayed:
- *
- *   Pan Rate                 = 200
- *   Frames Processed         = 1225
- *   Width/Height/Depth       = 640/1024/8
- *   Zoom Factor (x/y)        = 3.000000/3.000000
- *   Update Duration          = (avg. 0.010232) (max. 0.002685) (tot. 1.084589) seconds
- *     Update Pixels          = (avg. 3046) pixels/frame
- *     Update Rate            = (avg. 97.732909) (max. 372.439485) updates/second
- *     Net Update Rate        = (avg. 0.297724) (max. 0.457741) Mpex/second
- *   Pan Latency              = (avg. 0.008229) (max. 0.562326) seconds
- *   Total Frame Duration     = (avg. 0.008363) (max. 0.562485) (tot. 10.235981) seconds
- *   Frame Rate               = (avg. 119.578189) (max. 4854.369141) frames/second
- *   Scroll Delta (x)         = (avg. 0.000000) (tot. 0) lines
- *   Scroll Delta (y)         = (avg. 5.019608) (tot. 6144) lines
- *   Scroll Rate (x)          = (avg. 0.000000) lines/second
- *   Scroll Rate (y)          = (avg. 600.235657) lines/second
- *   Net Render Rate          = (avg. 8.685322) (max. 3181.359360) Mpex/second
- *
- * Pan Rate             - Displays the pan rate specified on the command line.
- *                        Only displayed if --timing-pan-rate is used.
- * Frames Processed     - Number of frames displayed
- * Width/Height/Depth   - The width/height/depth of magnifier display.
- * Zoom Factor (x/y)    - The zoom factor used.
- * Update Duration      - Average and total time spent updating new sections of the screen.
- *   Update Pixels      - Number of pixels updated
- *   Update Rate        - Number of updates possible per second.
- *   Net Update Rate    - Number of Megapixels updated per second.
- * Pan Latency          - Time spent from the beginning of the setROI request until GTK+
- *                        has finished processing the update (in GDK's
- *                        gdk_window_update_idle idle handler.
- * Total Frame Duration - Average and total time processing frames.
- * Frame Rate           - Number of frames per second.
- * Scroll Delta         - Average and total number of lines scrolled.
- * Scroll Rate          - Number of lines scrolled per second.  Note that this value
- *                        will be equal to the --timing-pan-rate value times the
- *                        y zoom factor.
- * Net Render Rate      - Number of Megapixels processed per second.
- *
- * When running with the --timing-output argument, similar information is displayed for
- * each frame plus timing information for the frame.  Update information is not
- * displayed for frames where updates do not occur.  Pan Rate, Frames Processed,
- * and Width/Height/Depth is not displayed for each frame, only when displaying
- * summary information.
- */
-
-typedef struct timing {
-	struct timeval process_start;
-	struct timeval process_end;
-	struct timeval frame_start;
-	struct timeval frame_end;
-	struct timeval scale_start;
-	struct timeval scale_end;
-	struct timeval idle_start;
-	struct timeval idle_end;
-	float scale_val;
-	float idle_val;
-	float frame_val;
-	int   dx;
-	int   dy;
-	float scale_total;
-	float idle_total;
-	float frame_total;
-	int   dx_total;
-	int   dy_total;
-	long  update_pixels_total;
-	long  num_scale_samples;
-	long  num_idle_samples;
-	long  num_frame_samples;
-	long  num_line_samples;
-	float last_frame_val;
-	int   last_dy;
-} TimingStruct;
-
 static float timing_scale_max  = 0;
 static float timing_idle_max   = 0;
 static float timing_frame_max  = 0;
@@ -193,7 +120,6 @@ static gboolean	reset_timing   = FALSE;
 static gboolean timing_test    = FALSE;
 
 static guint pending_idle_handler = 0;
-static TimingStruct mag_timing;
 static gboolean processing_updates = FALSE;
 static gboolean timing_start = FALSE;
 
@@ -220,6 +146,7 @@ static void zoom_region_recompute_exposed_viewport (ZoomRegion *zoom_region);
 static GdkRectangle zoom_region_rect_from_bounds (ZoomRegion *zoom_region,
 						  const GNOME_Magnifier_RectBounds *bounds);
 static ZoomRegionPixmapCreationError zoom_region_create_pixmap (ZoomRegion *zoom_region);
+static GdkRectangle zoom_region_update_pixmap (ZoomRegion *zoom_region, const GdkRectangle update_rect, GdkRectangle *paint_rect);
 
 void
 reset_timing_stats()
@@ -243,7 +170,7 @@ reset_timing_stats()
 	mag_timing.dy_total            = 0;
 	mag_timing.last_frame_val      = 0;
 	mag_timing.last_dy             = 0;
-	gettimeofday(&mag_timing.process_start, NULL);
+	g_timer_start (mag_timing.process);
 }
 
 /** DEBUG STUFF **/
@@ -261,29 +188,6 @@ _debug_announce_rect (char *msg, GdkRectangle rect)
 		 msg, rect.x, rect.y, rect.x + rect.width, rect.y + rect.height);
 }
 
-#ifdef TEST_XTST_CURSOR
-static Cursor *
-_get_x_cursors (Display *display)
-{
-	int n_x_font_cursors = XC_num_glyphs; 
-	int dummy;
-	if (!XTestQueryExtension (display, &dummy, &dummy,&dummy, &dummy))
-		fprintf (stderr, "WARNING: XTEST extension not available!\n");
-	if (x_cursors == NULL && display) {
-		int i;
-		x_cursors = g_new0 (Cursor, XC_num_glyphs);
-		for (i = 2; i < n_x_font_cursors; i += 2)
-		{
-			if (i == 30)
-				x_cursors[i] = XCreateFontCursor (display, XC_xterm);
-			else
-				x_cursors[i] = XCreateFontCursor (display, i);
-		}
-	}
-	return x_cursors;
-}
-#endif
-
 static gboolean
 _diff_pixbufs (const GdkPixbuf *a, const GdkPixbuf *b)
 {
@@ -640,35 +544,25 @@ static GdkRectangle
 zoom_region_clip_to_exposed_target (ZoomRegion *zoom_region,
 				    GdkRectangle area)
 {
-	GNOME_Magnifier_RectBounds onscreen_target;
-	onscreen_target.x1 = MAX (zoom_region->priv->exposed_bounds.x1 / zoom_region->xscale,
-				  zoom_region->priv->source_area.x1);
-	onscreen_target.y1 = MAX (zoom_region->priv->exposed_bounds.y1 / zoom_region->yscale,
-				  zoom_region->priv->source_area.y1);
-	onscreen_target.x2 = MIN (zoom_region->priv->exposed_bounds.x2 / zoom_region->xscale,
-				  zoom_region->priv->source_area.x2);
-	onscreen_target.y2 = MIN (zoom_region->priv->exposed_bounds.y2 / zoom_region->yscale,
-				  zoom_region->priv->source_area.y2);
+	GNOME_Magnifier_RectBounds onscreen_target, *source_area;
+	source_area = &zoom_region->priv->source_area;
+
+	onscreen_target.x1 = MAX (floor (zoom_region->priv->exposed_bounds.x1
+					 / zoom_region->xscale),
+					 source_area->x1);
+	onscreen_target.y1 = MAX (floor (zoom_region->priv->exposed_bounds.y1
+					 / zoom_region->yscale),
+					 source_area->y1);
+	onscreen_target.x2 = MIN (ceil (zoom_region->priv->exposed_bounds.x2
+					/ zoom_region->xscale),
+					source_area->x2);
+	onscreen_target.y2 = MIN (ceil (zoom_region->priv->exposed_bounds.y2
+					/ zoom_region->yscale),
+					source_area->y2);
 
 	return _rectangle_clip_to_bounds (area, &onscreen_target);
 }
 
-#ifdef NEEDED
-static GdkRectangle
-zoom_region_clip_to_exposed_bounds (ZoomRegion *zoom_region,
-				    GdkRectangle area)
-{
-	return _rectangle_clip_to_bounds (area, &zoom_region->priv->exposed_bounds);
-}
-
-static GdkRectangle
-zoom_region_clip_to_viewport (ZoomRegion *zoom_region,
-			      GdkRectangle area)
-{
-	return _rectangle_clip_to_bounds (area, &zoom_region->priv->exposed_viewport);
-}
-#endif
-
 static GdkRectangle
 zoom_region_clip_to_scaled_pixmap (ZoomRegion *zoom_region,
 				   GdkRectangle area)
@@ -714,8 +608,8 @@ zoom_region_source_rect_from_view_bounds (ZoomRegion *zoom_region,
 			       / zoom_region->xscale);
 	source_rect.y = floor ((view_bounds->y1 + zoom_region->priv->exposed_bounds.y1)
 				/ zoom_region->yscale);
-	source_rect.width = ceil ((view_bounds->x2 - view_bounds->x1) / zoom_region->xscale);
-	source_rect.height = ceil ((view_bounds->y2 - view_bounds->y1) / zoom_region->yscale);
+	source_rect.width = ceil ((view_bounds->x2 - view_bounds->x1) / zoom_region->xscale) + 1;
+	source_rect.height = ceil ((view_bounds->y2 - view_bounds->y1) / zoom_region->yscale) + 1;
 	return source_rect;
 }
 
@@ -742,8 +636,8 @@ zoom_region_source_rect_from_view_rect (ZoomRegion *zoom_region,
 			       / zoom_region->xscale);
 	source_rect.y = floor ((view_rect.y + zoom_region->priv->exposed_bounds.y1)
 				/ zoom_region->yscale);
-	source_rect.width = ceil (view_rect.width / zoom_region->xscale);
-	source_rect.height = ceil (view_rect.height / zoom_region->yscale);
+	source_rect.width = ceil (view_rect.width / zoom_region->xscale) + 1;
+	source_rect.height = ceil (view_rect.height / zoom_region->yscale) + 1;
 	return source_rect;
 }
 
@@ -861,9 +755,8 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 	GdkRectangle rect;
 	GdkDrawable *cursor;
 	GdkColor color = {0, 0, 0, 0};
-	GdkColor tmp_color;
 	int x_left_clip = 0, x_right_clip = 0, y_top_clip = 0, y_bottom_clip = 0;
-	int csize;
+	int csize = 0;
 	
 #ifdef ZOOM_REGION_DEBUG
 	g_assert (zoom_region->alive);
@@ -925,10 +818,11 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 	if (clip_rect) gdk_gc_set_clip_rectangle (zoom_region->priv->crosswire_gc, clip_rect);
 	else gdk_gc_set_clip_rectangle (zoom_region->priv->crosswire_gc, NULL);
 
-	if (magnifier->crosswire_clip &&
-	    (cursor = magnifier_get_cursor (magnifier)))
-	{
+	if ((cursor = magnifier_get_cursor (magnifier))) {
 		gdk_drawable_get_size (cursor, &csize, &csize);
+	}
+	if (magnifier->crosswire_clip)
+	{
 		y_top_clip = rect.y - magnifier->cursor_hotspot.y -
 			magnifier->crosswire_size;
 		y_bottom_clip = rect.y +
@@ -939,6 +833,7 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 		x_right_clip = rect.x +
 			(csize - magnifier->cursor_hotspot.x) +
 			magnifier->crosswire_size;
+
 	}
 	if (magnifier->crosswire_size == 1)
 	{
@@ -1090,7 +985,7 @@ zoom_region_coalesce_updates (ZoomRegion *zoom_region)
 		zoom_region->priv->q = NULL; /* just discard and update everything */
 		/* CAUTION: this can be an expensive operation! */
 		zoom_region_queue_update (zoom_region, zoom_region_rect_from_bounds
-			(zoom_region, &zoom_region->priv->source_area));
+					  (zoom_region, &zoom_region->priv->source_area));
 	}
 	else 
 
@@ -1169,8 +1064,8 @@ zoom_region_paint_pixmap (ZoomRegion *zoom_region,
 		gdk_draw_drawable (zoom_region->priv->w->window,
 				   zoom_region->priv->default_gc,
 				   zoom_region->priv->pixmap,
-				   area->x + zoom_region->priv->exposed_bounds.x1,
-				   area->y + zoom_region->priv->exposed_bounds.y1,
+				   area->x + zoom_region->priv->exposed_bounds.x1 - zoom_region->priv->source_area.x1 * zoom_region->xscale,
+				   area->y + zoom_region->priv->exposed_bounds.y1 - zoom_region->priv->source_area.y1 * zoom_region->yscale,
 				   area->x,
 				   area->y,
 				   area->width,
@@ -1279,8 +1174,9 @@ zoom_region_expose_handler (GtkWindow * w,
 	zoom_region_paint (zoom_region, &event->area);
 }
 
-static void zoom_region_update_cursor (ZoomRegion *zoom_region, int dx, int dy,
-				       GdkRectangle *clip_rect)
+static void
+zoom_region_update_cursor (ZoomRegion *zoom_region, int dx, int dy,
+			   GdkRectangle *clip_rect)
 {
 #ifdef ZOOM_REGION_DEBUG
 	g_assert (zoom_region->alive);
@@ -1331,12 +1227,13 @@ zoom_region_calculate_scroll_rects (ZoomRegion *zoom_region,
 		DBG(fprintf (stderr, "deltas too big to scroll\n"));
 		retval = FALSE;
 	}
-	
-	scroll_rect->x = MAX (0, dx);
-	scroll_rect->y = MAX (0, dy);
-	scroll_rect->width = MIN (rect.width + dx, rect.width - dx);
-	scroll_rect->height = MIN (rect.height + dy, rect.height - dy);
-	
+	else {
+	    scroll_rect->x = MAX (0, dx);
+	    scroll_rect->y = MAX (0, dy);
+	    scroll_rect->width = MIN (rect.width + dx, rect.width - dx);
+	    scroll_rect->height = MIN (rect.height + dy, rect.height - dy);
+	}
+
 	expose_rect_h->x = 0;
 	expose_rect_h->y = (scroll_rect->y == 0) ? scroll_rect->height : 0;
 	expose_rect_h->width = rect.width;
@@ -1399,9 +1296,7 @@ zoom_region_scroll_smooth (ZoomRegion *zoom_region, int dx, int dy,
 	gdk_drawable_get_size (GDK_DRAWABLE (window),
 			       &window_rect.width, &window_rect.height);
 	gdk_window_begin_paint_rect (window, &window_rect);
-	gdk_window_invalidate_rect (window, scroll_rect, FALSE);
-	gdk_window_invalidate_rect (window, expose_rect_h, FALSE);
-	gdk_window_invalidate_rect (window, expose_rect_v, FALSE);
+	gdk_window_invalidate_rect (window, &window_rect, FALSE);
 	gdk_window_process_updates (window, FALSE); 
 	gdk_window_end_paint (window);
 }
@@ -1439,30 +1334,24 @@ zoom_region_scroll (ZoomRegion *zoom_region, int dx, int dy)
 							 &scroll_rect,
 							 &expose_rect_h,
 							 &expose_rect_v);
-	if (zoom_region->smooth_scroll_policy > GNOME_Magnifier_ZoomRegion_SCROLL_FAST)
-	{
-		zoom_region_scroll_smooth (zoom_region,
-					   dx, dy, &scroll_rect,
-					   &expose_rect_h, &expose_rect_v);
+	
+	if (can_scroll) {
+		zoom_region_update_pixmap (zoom_region, zoom_region_source_rect_from_view_rect (zoom_region, expose_rect_h), NULL);
+		zoom_region_update_pixmap (zoom_region, zoom_region_source_rect_from_view_rect (zoom_region, expose_rect_v), NULL);
+
+		if (zoom_region->smooth_scroll_policy > GNOME_Magnifier_ZoomRegion_SCROLL_FAST) {
+			zoom_region_scroll_smooth (zoom_region, dx, dy,
+						   &scroll_rect,
+						   &expose_rect_h,
+						   &expose_rect_v);
+		} else {
+			zoom_region_scroll_fast (zoom_region, dx, dy,
+						 &scroll_rect,
+						 &expose_rect_h,
+						 &expose_rect_v);
+		}
 	} else {
-		zoom_region_scroll_fast (zoom_region,
-					 dx, dy, &scroll_rect,
-					 &expose_rect_h, &expose_rect_v);
-	}
-	if (!can_scroll)	
-	{
-		zoom_region_queue_update (zoom_region,
-					  zoom_region_source_rect_from_view_rect (zoom_region,
-										  scroll_rect));
-	}
-	else
-	{
-		zoom_region_queue_update (zoom_region,
-					  zoom_region_source_rect_from_view_rect (zoom_region,
-										  expose_rect_h));
-		zoom_region_queue_update (zoom_region,
-					  zoom_region_source_rect_from_view_rect (zoom_region,
-										  expose_rect_v));
+		zoom_region_queue_update (zoom_region, zoom_region_source_rect_from_view_rect (zoom_region, scroll_rect));
 	}
 }
 
@@ -1498,20 +1387,6 @@ zoom_region_set_cursor_pos (ZoomRegion *zoom_region, int x, int y)
 	}
 }
 
-#ifdef TEST_XTST_CURSOR
-static GdkWindow *
-zoom_region_source_root (ZoomRegion *zoom_region)
-{
-	GdkScreen *source_screen = NULL;
-	if (zoom_region->priv->source_drawable)
-		source_screen =
-			gdk_drawable_get_screen (zoom_region->priv->source_drawable);
-	if (source_screen)
-		return gdk_screen_get_root_window (source_screen);
-	else return NULL;
-}
-#endif
-
 static gboolean
 zoom_region_update_pointer (ZoomRegion *zoom_region, gboolean draw_cursor)
 {
@@ -1625,31 +1500,55 @@ zoom_region_moveto (ZoomRegion *zoom_region,
 	}
 }
 
-#define invert_rgb(a) \
-(~(a & 0x00FFFFFF) | (a & 0xFF000000))
-
-#define GET_PIXEL(a,i,j,s,b) \
-(*(guint32 *)(memcpy (b,(a) + ((j) * s + (i) * pixel_size_t), pixel_size_t)))
-
-#define PUT_PIXEL(a,i,j,s,b) \
-(memcpy (a + ((j) * s + (i) * pixel_size_t), &(b), pixel_size_t))
-
+/*
+ * Process that must be made in-line in the current pixbuf.
+ */
 static void
-_zoom_region_invert_pixbuf (GdkPixbuf *pixbuf)
+zoom_region_process_pixbuf (ZoomRegion *zoom_region, GdkPixbuf *pixbuf)
 {
 	int rowstride = gdk_pixbuf_get_rowstride (pixbuf);
-	int i, j;
+	int i, j, t;
 	int w = gdk_pixbuf_get_width (pixbuf);
 	int h = gdk_pixbuf_get_height (pixbuf);
+	int n_channels = gdk_pixbuf_get_n_channels (pixbuf);
 	guchar *pixels = gdk_pixbuf_get_pixels (pixbuf);
-	guint32 pixval = 0;
-	size_t pixel_size_t = 3; /* FIXME: invalid assumption ? */
+	guchar *pixels_row;
+
+	gboolean manipulate_contrast = FALSE;
+
+	if (zoom_region->contrast_r != 1 || zoom_region->contrast_g != 1 ||
+	    zoom_region->contrast_b != 1) {
+		manipulate_contrast = TRUE;
+	}
+
+	if (!(manipulate_contrast || zoom_region->invert))
+		return;
+
+#define CLAMP_UCHAR(v) (t = (v), CLAMP (t, 0, 255))
 
 	for (j = 0; j < h; ++j) {
+		pixels_row = pixels;
 		for (i = 0; i < w; ++i) {
-			pixval = invert_rgb (GET_PIXEL (pixels, i, j, rowstride, &pixval));
-			PUT_PIXEL (pixels, i, j, rowstride, pixval);
+			if (manipulate_contrast) {
+				// Set the RED contrast
+				pixels_row[0] = CLAMP_UCHAR (pixels_row[0] * zoom_region->contrast_r);
+
+				// Set the GREEN contrast
+				pixels_row[1] = CLAMP_UCHAR (pixels_row[1] * zoom_region->contrast_g);
+
+				// Set the BLUE contrast
+				pixels_row[2] = CLAMP_UCHAR (pixels_row[2] * zoom_region->contrast_b);
+			}
+
+			if (zoom_region->invert) {
+				pixels_row[0] = ~(pixels_row[0]);
+				pixels_row[1] = ~(pixels_row[1]);
+				pixels_row[2] = ~(pixels_row[2]);
+			}
+			
+			pixels_row += n_channels;
 		}
+		pixels += rowstride;
 	}
 }
 
@@ -1722,7 +1621,6 @@ zoom_region_get_source_subwindow (ZoomRegion *zoom_region,
 		}
 		else
 		{
-			Magnifier *magnifier = (Magnifier *) zoom_region->priv->parent;
 			zoom_region->priv->source_drawable = gdk_screen_get_root_window (
 				gdk_display_get_screen (
 					magnifier->source_display,
@@ -1774,15 +1672,10 @@ zoom_region_get_source_subwindow (ZoomRegion *zoom_region,
 	return subimage;
 }
 
-/**
- * zoom_region_update:
- *
- * @rect: a RectBounds structure indicating the source area to update,
- *        in the source coordinate system.
- **/
-static void
-zoom_region_update (ZoomRegion *zoom_region,
-		    const GdkRectangle update_rect)
+static GdkRectangle
+zoom_region_update_pixmap (ZoomRegion *zoom_region,
+			   const GdkRectangle update_rect,
+			   GdkRectangle *p_rect)
 {
 	GdkPixbuf *subimage;
 	GdkRectangle source_rect;
@@ -1798,20 +1691,23 @@ zoom_region_update (ZoomRegion *zoom_region,
 
 	subimage = zoom_region_get_source_subwindow (zoom_region, source_rect);
 
-	if (subimage && zoom_region->priv->w && zoom_region->priv->w->window)
+	if (subimage)
 	{
 		GdkRectangle paint_rect;
-		gettimeofday(&mag_timing.scale_start, NULL);
+		g_timer_start (mag_timing.scale);
 		DEBUG_RECT ("source rect", source_rect);
 		paint_rect = zoom_region_view_rect_from_source_rect (zoom_region, source_rect);
+		if (p_rect) {
+			*p_rect = paint_rect;
+		}
 		/* paint_rect = zoom_region_clip_to_scaled_pixmap (zoom_region, paint_rect); */
 		DEBUG_RECT ("paint rect", paint_rect);
-		if (zoom_region->invert)
-			_zoom_region_invert_pixbuf (subimage);
+
+		zoom_region_process_pixbuf (zoom_region, subimage);
 
 		/** 
-		 *   XXX: We seem to be breaking with the original intention here, which was to 
-		 *   keep a fullscreen scaled pixbuf in-sync.
+		 * XXX: We seem to be breaking with the original intention 
+		 * here, which was to keep a fullscreen scaled pixbuf in-sync.
 		 **/
 		gdk_pixbuf_scale (subimage,
 				  zoom_region->priv->scaled_pixbuf,
@@ -1837,8 +1733,8 @@ zoom_region_update (ZoomRegion *zoom_region,
 				     zoom_region->priv->scaled_pixbuf,
 				     0,
 				     0,
-				     paint_rect.x + zoom_region->priv->exposed_bounds.x1,
-				     paint_rect.y + zoom_region->priv->exposed_bounds.y1,
+				     paint_rect.x + zoom_region->priv->exposed_bounds.x1 - zoom_region->priv->source_area.x1 * zoom_region->xscale,
+				     paint_rect.y + zoom_region->priv->exposed_bounds.y1 - zoom_region->priv->source_area.y1 * zoom_region->yscale,
 				     paint_rect.width,
 				     paint_rect.height,
 				     GDK_RGB_DITHER_NONE,
@@ -1863,16 +1759,39 @@ zoom_region_update (ZoomRegion *zoom_region,
 		if (magnifier_error_check ())
 			g_warning ("Could not render scaled image to drawable; out of memory!\n");
 		g_object_unref (subimage);
-		gdk_window_begin_paint_rect (zoom_region->priv->w->window, &paint_rect);
-		zoom_region_paint (zoom_region, &paint_rect);
-		gdk_window_end_paint (zoom_region->priv->w->window);
-   		gettimeofday(&mag_timing.scale_end, NULL);
+
+		g_timer_stop (mag_timing.scale);
+	}
+	return source_rect;
+}
+
+/**
+ * zoom_region_update:
+ *
+ * @rect: a RectBounds structure indicating the source area to update,
+ *        in the source coordinate system.
+ **/
+static void
+zoom_region_update (ZoomRegion *zoom_region,
+		    const GdkRectangle update_rect)
+{
+	GdkRectangle paint_rect = {0, 0, 0, 0};
+	if (zoom_region->priv->w && zoom_region->priv->w->window) {
+		GdkRectangle source_rect = zoom_region_update_pixmap (zoom_region, update_rect, &paint_rect);
+		if (paint_rect.x != 0 || paint_rect.y != 0 ||
+		    paint_rect.width != 0 || paint_rect.height != 0) {
+			gdk_window_begin_paint_rect (zoom_region->priv->w->window, &paint_rect);
+			zoom_region_paint (zoom_region, &paint_rect);
+			gdk_window_end_paint (zoom_region->priv->w->window);
+		}
 		if (timing_test) {
 			mag_timing.num_scale_samples++;
+			
+			gulong microseconds;
+
 			mag_timing.scale_val =
-				(mag_timing.scale_end.tv_sec - mag_timing.scale_start.tv_sec) +
-				((float)(mag_timing.scale_end.tv_usec - mag_timing.scale_start.tv_usec) /
-				1000000.0);
+			        g_timer_elapsed (mag_timing.scale,
+						 &microseconds);
 			mag_timing.scale_total += mag_timing.scale_val;
 
 			if (mag_timing.scale_val != 0 && (timing_scale_max == 0 ||
@@ -1898,8 +1817,7 @@ zoom_region_update (ZoomRegion *zoom_region,
 			}
 		}
 	} else {
-		if (subimage)
-			fprintf (stderr, "update on uninitialized zoom region!\n");
+		fprintf (stderr, "update on uninitialized zoom region!\n");
 	}
 }
 
@@ -2045,10 +1963,12 @@ zoom_region_time_frame(ZoomRegion *zoom_region, Magnifier *magnifier)
 	int height = magnifier->target_bounds.y2 - magnifier->target_bounds.y1;
 
 	mag_timing.num_frame_samples++;
-	gettimeofday(&mag_timing.frame_end, NULL);
-	mag_timing.frame_val = (mag_timing.frame_end.tv_sec - mag_timing.frame_start.tv_sec) +
-		((float)(mag_timing.frame_end.tv_usec - mag_timing.frame_start.tv_usec) /
-		1000000.0);
+	g_timer_stop (mag_timing.frame);
+
+	gulong microseconds;
+
+	mag_timing.frame_val = g_timer_elapsed (mag_timing.frame,
+						&microseconds);
 
 	mag_timing.frame_total += mag_timing.frame_val;
 	if (mag_timing.frame_val > timing_frame_max)
@@ -2096,57 +2016,6 @@ zoom_region_time_frame(ZoomRegion *zoom_region, Magnifier *magnifier)
 	}
 }
 
-void
-exit_report(ZoomRegion *zoom_region)
-{
-	float frame_avg;
-	float x_scroll_incr, y_scroll_incr;
-	int width, height, x, y;
-
-	if (timing_test) {
-		width = (zoom_region->priv->exposed_viewport.x2 -
-			zoom_region->priv->exposed_viewport.x1) / zoom_region->xscale;
-		height = (zoom_region->priv->exposed_viewport.y2 -
-			zoom_region->priv->exposed_viewport.y1) / zoom_region->yscale;
-
-		frame_avg = mag_timing.frame_total / mag_timing.num_frame_samples;
-
-		x_scroll_incr = (float)mag_timing.dx_total / (float)mag_timing.num_line_samples;
-		y_scroll_incr = (float)mag_timing.dy_total / (float)mag_timing.num_line_samples;
-
-		gdk_drawable_get_size (GDK_DRAWABLE (zoom_region->priv->w->window),
-			&x, &y);
-
-		fprintf(stderr, "\nFinal Report:\n\n");
-
-		fprintf(stderr, "  Frames Processed         = %ld\n", 
-			mag_timing.num_frame_samples + 1);
-		fprintf(stderr, "  Width/Height/Depth       = %d/%d/%d\n", x, y,
-			gdk_drawable_get_depth (zoom_region->priv->w->window));
-		fprintf(stderr, "  Zoom Factor (x/y)        = %f/%f\n", zoom_region->xscale,
-			zoom_region->yscale);
-		fprintf(stderr, "  Pan Latency              = (avg. %f) (max. %f) seconds\n",
-			(mag_timing.idle_total / mag_timing.num_idle_samples), timing_idle_max);
-		fprintf(stderr, "  Total Frame Duration     = (avg. %f) (max. %f) (tot. %f) seconds\n",
-			frame_avg, timing_frame_max, mag_timing.frame_total);
-		fprintf(stderr, "  Frame Rate               = (avg. %f) (max. %f) frames/second\n",
-			1.0 / (mag_timing.frame_total / mag_timing.num_frame_samples), cps_max);
-		fprintf(stderr, "  Scroll Delta (x)         = (avg. %f) (tot. %d) lines\n",
-			x_scroll_incr, mag_timing.dx_total);
-		fprintf(stderr, "  Scroll Delta (y)         = (avg. %f) (tot. %d) lines\n",
-			y_scroll_incr, mag_timing.dy_total);
-		fprintf(stderr, "  Scroll Rate (x)          = (avg. %f) lines/second\n",
-			x_scroll_incr / frame_avg);
-		fprintf(stderr, "  Scroll Rate (y)          = (avg. %f) lines/second\n",
-			y_scroll_incr / frame_avg);
-
-		fprintf(stderr, "  Net Render Rate          = (avg. %f) (max. %f) Mpex/second\n\n",
-			(height * width *
-			((float)mag_timing.num_frame_samples / (float)mag_timing.frame_total)) / 1000000.0,
-			nrr_max / 1000000.0);
-	}
-}
-
 static void
 zoom_region_sync (ZoomRegion *zoom_region)
 {
@@ -2161,14 +2030,15 @@ gdk_timing_idle (gpointer data)
 
 	/* Now update has finished, reset processing_updates */
 	processing_updates = FALSE;
-	gettimeofday(&mag_timing.idle_end, NULL);
+	g_timer_stop (mag_timing.idle);
 
 	if (timing_test) {
 		mag_timing.num_idle_samples++;
-		mag_timing.idle_val =
-			(mag_timing.idle_end.tv_sec - mag_timing.idle_start.tv_sec) +
-			((float)(mag_timing.idle_end.tv_usec - mag_timing.idle_start.tv_usec) /
-			1000000.0);
+
+		gulong microseconds;
+
+		mag_timing.idle_val = g_timer_elapsed (mag_timing.idle,
+						       &microseconds);
 		mag_timing.idle_total += mag_timing.idle_val;
 
 		if (mag_timing.idle_val > timing_idle_max)
@@ -2195,7 +2065,7 @@ zoom_region_align (ZoomRegion *zoom_region)
 		zoom_region_time_frame(zoom_region, magnifier);
 
 	if (timing_test) {
-		gettimeofday(&mag_timing.frame_start, NULL);
+		g_timer_start (mag_timing.frame);
 
 		if (zoom_region->timing_output) {
 			gint x, y;
@@ -2217,12 +2087,12 @@ zoom_region_align (ZoomRegion *zoom_region)
 		 * until we have processed an entire frame.
 		 */
 		if (!timing_start)
-			gettimeofday(&mag_timing.process_start, NULL);
+		        g_timer_start (mag_timing.process);
 
 		timing_start = TRUE;
 	}
 
-	gettimeofday(&mag_timing.idle_start, NULL);
+	g_timer_start (mag_timing.idle);
 
 	/*
 	 * zoom_region_align calls
@@ -2294,6 +2164,9 @@ zoom_region_set_viewport (ZoomRegion *zoom_region,
 	if (!zoom_region->priv->w) {
 		zoom_region_init_window (zoom_region);
 	} else {
+	        CORBA_any *any;
+		CORBA_Environment ev;
+		Bonobo_PropertyBag properties;
 		Magnifier *magnifier = (Magnifier *) zoom_region->priv->parent;
 		GtkFixed *fixed = GTK_FIXED (magnifier->priv->canvas);
 		gtk_fixed_move (fixed,
@@ -2305,6 +2178,29 @@ zoom_region_set_viewport (ZoomRegion *zoom_region,
 					     zoom_region->viewport.x1,
 					     zoom_region->viewport.y2 -
 					     zoom_region->viewport.y1);
+		CORBA_exception_init (&ev);
+		properties = 
+			GNOME_Magnifier_Magnifier_getProperties(
+				BONOBO_OBJREF (
+					(Magnifier *) zoom_region->priv->parent), &ev);
+		if (!BONOBO_EX (&ev))
+			any = Bonobo_PropertyBag_getValue (
+				properties, "source-display-bounds", &ev);
+		if (!BONOBO_EX (&ev))
+			zoom_region->priv->source_area =
+				*((GNOME_Magnifier_RectBounds *) any->_value);
+		if (zoom_region->priv->pixmap) 
+			g_object_unref (zoom_region->priv->pixmap);
+		zoom_region_create_pixmap (zoom_region);
+		if (zoom_region->priv->scaled_pixbuf)
+		        g_object_unref (zoom_region->priv->scaled_pixbuf);
+
+		zoom_region->priv->scaled_pixbuf = 
+		  gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8,
+				  (zoom_region->priv->source_area.x2 -
+				   zoom_region->priv->source_area.x1) * zoom_region->xscale + 1,
+				  (zoom_region->priv->source_area.y2 -
+				   zoom_region->priv->source_area.y1) * zoom_region->yscale + 1);
 	}
 	zoom_region_queue_update (zoom_region,
 				  zoom_region_source_rect_from_view_bounds (
@@ -2341,8 +2237,14 @@ zoom_region_get_property (BonoboPropertyBag *bag,
 	case ZOOM_REGION_SMOOTHING_PROP:
 		BONOBO_ARG_SET_STRING (arg, zoom_region->smoothing);
 		break;
-	case ZOOM_REGION_CONTRAST_PROP:
-		BONOBO_ARG_SET_FLOAT (arg, zoom_region->contrast);
+	case ZOOM_REGION_CONTRASTR_PROP:
+		BONOBO_ARG_SET_FLOAT (arg, zoom_region->contrast_r);
+		break;
+	case ZOOM_REGION_CONTRASTG_PROP:
+		BONOBO_ARG_SET_FLOAT (arg, zoom_region->contrast_g);
+		break;
+	case ZOOM_REGION_CONTRASTB_PROP:
+		BONOBO_ARG_SET_FLOAT (arg, zoom_region->contrast_b);
 		break;
 	case ZOOM_REGION_XSCALE_PROP:
 		BONOBO_ARG_SET_FLOAT (arg, zoom_region->xscale);
@@ -2427,8 +2329,16 @@ zoom_region_set_property (BonoboPropertyBag *bag,
 			g_object_unref (zoom_region->priv->source_drawable);
 		zoom_region_update_current (zoom_region);
 		break;
-	case ZOOM_REGION_CONTRAST_PROP:
-		zoom_region->contrast = BONOBO_ARG_GET_FLOAT (arg);
+	case ZOOM_REGION_CONTRASTR_PROP:
+		zoom_region->contrast_r = BONOBO_ARG_GET_FLOAT (arg);
+		zoom_region_update_current (zoom_region);
+		break;
+	case ZOOM_REGION_CONTRASTG_PROP:
+		zoom_region->contrast_g = BONOBO_ARG_GET_FLOAT (arg);
+		zoom_region_update_current (zoom_region);
+		break;
+	case ZOOM_REGION_CONTRASTB_PROP:
+		zoom_region->contrast_b = BONOBO_ARG_GET_FLOAT (arg);
 		zoom_region_update_current (zoom_region);
 		break;
 	case ZOOM_REGION_XSCALE_PROP:
@@ -2454,12 +2364,12 @@ zoom_region_set_property (BonoboPropertyBag *bag,
 		break;
 	case ZOOM_REGION_XALIGN_PROP:
 		zoom_region->x_align_policy = BONOBO_ARG_GET_INT (arg);
-		zoom_region_update_current (zoom_region);
+		zoom_region_align (zoom_region);
 		break;
 	case ZOOM_REGION_YALIGN_PROP:
 		/* TODO: enums here */
 		zoom_region->y_align_policy = BONOBO_ARG_GET_INT (arg);
-		zoom_region_update_current (zoom_region);
+		zoom_region_align (zoom_region);
 		break;
 	case ZOOM_REGION_VIEWPORT_PROP:
 		bounds = BONOBO_ARG_GET_GENERAL (arg,
@@ -2526,10 +2436,11 @@ zoom_region_pan_test (gpointer data)
 	roi.x1 = zoom_region->roi.x1;
 	roi.x2 = zoom_region->roi.x2;
 
-	gettimeofday(&mag_timing.process_end, NULL);
-	total_time = (mag_timing.process_end.tv_sec - mag_timing.process_start.tv_sec) +
-		((float)(mag_timing.process_end.tv_usec - mag_timing.process_start.tv_usec) /
-		1000000.0);
+	g_timer_stop (mag_timing.process);
+
+	gulong microseconds;
+
+	total_time = g_timer_elapsed (mag_timing.process, &microseconds);
 
 	if (mag_timing.frame_total != 0.0)
 		pixels_at_speed = total_time * zoom_region->timing_pan_rate;
@@ -2589,6 +2500,53 @@ zoom_region_pan_test (gpointer data)
 	return FALSE;
 }
 
+static void
+impl_zoom_region_set_contrast (PortableServer_Servant servant,
+			       const CORBA_float R,
+			       const CORBA_float G,
+			       const CORBA_float B,
+			       CORBA_Environment *ev)
+{
+	ZoomRegion *zoom_region =
+		ZOOM_REGION (bonobo_object_from_servant (servant));
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
+	DBG (fprintf (stderr, "Set contrast: \t%f,%f %f\n", R, G, B));
+
+	/* if the contrast values are the same, this is a NOP */
+	if (zoom_region->contrast_r == R &&
+	    zoom_region->contrast_g == G &&
+	    zoom_region->contrast_b == B)
+		return;
+
+	zoom_region->contrast_r = R;
+	zoom_region->contrast_g = G;
+	zoom_region->contrast_b = B;
+
+	zoom_region_update_current (zoom_region);
+}
+
+static void
+impl_zoom_region_get_contrast (PortableServer_Servant servant,
+			       CORBA_float *R,
+			       CORBA_float *G,
+			       CORBA_float *B,
+			       CORBA_Environment *ev)
+{
+	ZoomRegion *zoom_region =
+		ZOOM_REGION (bonobo_object_from_servant (servant));
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
+
+	*R = zoom_region->contrast_r;
+	*G = zoom_region->contrast_g;
+	*B = zoom_region->contrast_b;
+}
+
 static void
 impl_zoom_region_set_roi (PortableServer_Servant servant,
 			  const GNOME_Magnifier_RectBounds *bounds,
@@ -2868,6 +2826,8 @@ zoom_region_class_init (ZoomRegionClass *klass)
 	epv->getROI = impl_zoom_region_get_roi;
 	epv->moveResize = impl_zoom_region_move_resize;
 	epv->dispose = impl_zoom_region_dispose;
+	epv->setContrast = impl_zoom_region_set_contrast;
+	epv->getContrast = impl_zoom_region_get_contrast;
 
 	reset_timing_stats();
 #ifdef DEBUG_CLIENT_CALLS
@@ -2955,15 +2915,41 @@ zoom_region_properties_init (ZoomRegion *zoom_region)
 	BONOBO_ARG_SET_FLOAT (def, 1.0);
 
 	bonobo_property_bag_add (zoom_region->properties,
-				 "contrast",
-				 ZOOM_REGION_CONTRAST_PROP,
+				 "red-contrast",
+				 ZOOM_REGION_CONTRASTR_PROP,
 				 BONOBO_ARG_FLOAT,
 				 def,
-				 "image contrast ratio",
+				 "red image contrast ratio",
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
-
 	bonobo_arg_release (def);
+
+	def = bonobo_arg_new (BONOBO_ARG_FLOAT);
+	BONOBO_ARG_SET_FLOAT (def, 1.0);
+
+	bonobo_property_bag_add (zoom_region->properties,
+				 "green-contrast",
+				 ZOOM_REGION_CONTRASTG_PROP,
+				 BONOBO_ARG_FLOAT,
+				 def,
+				 "green image contrast ratio",
+				 Bonobo_PROPERTY_READABLE |
+				 Bonobo_PROPERTY_WRITEABLE);
+	bonobo_arg_release (def);
+
+	def = bonobo_arg_new (BONOBO_ARG_FLOAT);
+	BONOBO_ARG_SET_FLOAT (def, 1.0);
+
+	bonobo_property_bag_add (zoom_region->properties,
+				 "blue-contrast",
+				 ZOOM_REGION_CONTRASTB_PROP,
+				 BONOBO_ARG_FLOAT,
+				 def,
+				 "blue image contrast ratio",
+				 Bonobo_PROPERTY_READABLE |
+				 Bonobo_PROPERTY_WRITEABLE);
+	bonobo_arg_release (def);
+
 	def = bonobo_arg_new (BONOBO_ARG_FLOAT);
 	BONOBO_ARG_SET_FLOAT (def, 2.0);
 
@@ -3147,6 +3133,9 @@ zoom_region_init (ZoomRegion *zoom_region)
 	zoom_region_properties_init (zoom_region);
 	zoom_region->smooth_scroll_policy =
 		GNOME_Magnifier_ZoomRegion_SCROLL_SMOOTH;
+	zoom_region->contrast_r = 1.0;
+	zoom_region->contrast_g = 1.0;
+	zoom_region->contrast_b = 1.0;
 	zoom_region->invert = FALSE;
 	zoom_region->cache_source = FALSE;
 	zoom_region->border_size = 0;
